import gi
gi.require_version("Gst", "1.0")
from gi.repository import Gst


class Camera():
	def __init__(self, device, camera_number, width, height, framerate, bitrate):
		Gst.init(None)

		# set your device camera, you should know this device name and camera number in this device and your resolution
		# the device name is below. camera number, you can check this number to type in terminal(ls /dev/video*, v4l2-ctl --list-formats-ext --device={camera number})
		if device == "RPi 4":
			self.cmd = f'v4l2src device=/dev/video{camera_number} ! video/x-raw, width={width}, height={height}, framerate={framerate}/1 ! videoconvert ! v4l2h264enc extra-controls="controls, video_bitrate={bitrate}, h264_livel=1, h264_profile=9, h264_i_frame_period={framerate}" ! video/x-h264, level=(string)3.1, profile=baseline, alignment=au, stream-format=byte-stream ! h264parse config-interval=1 ! queue leaky=2 ! appsink name=sink drop=true sync=false max-buffers=3 emit-signals=true'
			self.mime = f"video/h264;width={width};height={height};framerate={framerate};bitrate={bitrate};codecs=avc1.42002A"
		elif device == "RPi Zero 2W":
			self.cmd = f'v4l2src device=/dev/video{camera_number} ! video/x-raw, width={width}, height={height}, framerate={framerate}/1 ! videoflip method=rotate-180 ! videoconvert ! v4l2h264enc extra-controls="controls, video_bitrate={bitrate}, h264_livel=1, h264_profile=9, h264_i_frame_period={framerate}" ! video/x-h264, level=(string)3.1, profile=baseline, alignment=au, stream-format=byte-stream ! h264parse config-interval=1 ! queue leaky=2 ! appsink name=sink drop=true sync=false max-buffers=3 emit-signals=true'
			self.mime = f"video/h264;width={width};height={height};framerate={framerate};bitrate={bitrate};codecs=avc1.42002A"
		elif device == "RPi Zero":
			self.cmd = f'v4l2src device=/dev/video{camera_number} ! video/x-raw, width={width}, height={height}, framerate={framerate}/1 ! videoflip method=rotate-180 ! videoconvert ! v4l2h264enc extra-controls="controls, video_bitrate={bitrate}, h264_livel=1, h264_profile=9, h264_i_frame_period={framerate}" ! video/x-h264, level=(string)3.1, profile=baseline, alignment=au, stream-format=byte-stream ! h264parse config-interval=1 ! queue leaky=2 ! appsink name=sink drop=true sync=false max-buffers=3 emit-signals=true'
			self.mime = f"video/h264;width={width};height={height};framerate={framerate};bitrate={bitrate};codecs=avc1.42002A"
		elif device == "Jetson Nano CSI":
			self.cmd = f"nvarguscamerasrc sensor_id={camera_number} ! video/x-raw(memory:NVMM), width={width}, height={height}, framerate={framerate}/1 ! nvvidconv flip-method=2 ! nvv4l2h264enc bitrate={bitrate} insert-sps-pps=true idrinterval=30 iframeinterval=30 ! video/x-h264, alignment=au, stream-format=byte-stream ! h264parse config-interval=1 ! queue leaky=2 ! appsink name=sink drop=true sync=false max-buffers=3 emit-signals=true"
			self.mime = f"video/h264;width={width};height={height};framerate={framerate};bitrate={bitrate};codecs=avc1.42002A"
		elif device == "Jetson Nano USB":
			self.cmd = f"v4l2src device=/dev/video{camera_number} ! video/x-raw, width={width}, height={height}, framerate={framerate}/1 ! videoconvert ! nvvidconv ! nvv4l2h264enc bitrate={bitrate} insert-sps-pps=true idrinterval=30 iframeinterval=30 ! video/x-h264, alignment=au, stream-format=byte-stream ! h264parse config-interval=1 ! queue leaky=2 ! appsink name=sink drop=true sync=false max-buffers=3 emit-signals=true"
			self.mime = f"video/h264;width={width};height={height};framerate={framerate};bitrate={bitrate};codecs=avc1.42002A"
		elif device == "Jetson Orin Nano":
			self.cmd = f"v4l2src device=/dev/video{camera_number} ! video/x-raw, width={width}, height={height}, framerate={framerate}/1 ! videoconvert ! x264enc bitrate={bitrate} key-int-max=30 speed-preset=1 tune=zerolatency ! video/x-h264, alignment=au, stream-format=byte-stream ! h264parse config-interval=1 ! queue leaky=2 ! appsink name=sink drop=true sync=false max-buffers=3 emit-signals=true"
			self.mime = f"video/h264;width={width};height={height};framerate={framerate};bitrate={bitrate}000;codecs=avc1.42002A"
		elif device == "Jetson Orin NX":
			self.cmd = f"v4l2src device=/dev/video{camera_number} ! video/x-raw, width={width}, height={height}, framerate={framerate}/1 ! videoconvert ! nvvidconv ! nvv4l2h265enc bitrate={bitrate} insert-sps-pps=true idrinterval=30 iframeinterval=30 ! video/x-h265, alignment=au, stream-format=byte-stream ! h265parse config-interval=1 ! queue leaky=2 ! appsink name=sink drop=true sync=false max-buffers=3 emit-signals=true"
			self.mime = f"video/h265;width={width};height={height};framerate={framerate};bitrate={bitrate};codecs=hvc1.1.6.L123.00"

		# set GStreamer Pipeline
		self.pipeline = Gst.parse_launch(self.cmd)
		self.sink = self.pipeline.get_by_name("sink")
		self.pipeline.set_state(Gst.State.PLAYING)

	def getImage(self):
		sample = self.sink.emit("pull-sample")
		if sample:
			buf = sample.get_buffer()
			image = buf.extract_dup(0, buf.get_size())
			print("length: ", len(image))
			return image
		else:
			return False



