#include "CoviewRobot.hpp"


std::vector<uint8_t> pingTime(){
	auto ping_now = std::chrono::steady_clock::now();
	auto millis_ping = std::chrono::duration_cast<std::chrono::milliseconds>(ping_now.time_since_epoch()).count();

	std::vector<uint8_t> result_ping;
	result_ping.resize(4);
	result_ping[0] = (millis_ping >> 24) & 0xFF;
	result_ping[1] = (millis_ping >> 16) & 0xFF;
	result_ping[2] = (millis_ping >> 8) & 0xFF;
	result_ping[3] = millis_ping & 0xFF;

	return result_ping;
}

CoviewRobot::CoviewRobot(int w, int h, int n, int b) : width(w), height(h), num(n), bitrate(b)
{
	createCSV();
	initPing();
	initWebSocket();

	while (true){
		if (connection){
			std::this_thread::sleep_for(std::chrono::seconds(1));
			timeout++;

			bps_ = static_cast<double>(bps) / 1'000'000;
			bps = 0;
			if (err_count_flag == 0){
				err_count_flag = 1;
			}

			getProcessUsage("test");
			getProcessUsage("test_sub");

			if (timeout > 600) {
				timeout = 0;
				if (file.is_open()) {
					file.close();
				}
			}
		}
	}
};

void CoviewRobot::createCSV(){
	std::string filename = "./output.csv";

	file.open(filename, std::ios::app);

	if (!file.is_open()) {
		std::cerr << "Unable to open file" << std::endl;
		return;
	}

	if (file.tellp() == 0) {
		writeRow({"ERROR Count", "Ping Delay", "Sent Time", "Receive Time", "Latency", "Length", "Data Size", "BandWidth", "CPU Sub", "Memory Sub", "CPU Pub", "Memory Pub"});
	}
}

std::string CoviewRobot::exec(const char* cmd) {
	std::array<char, 128> buffer;
	std::string result;
	std::shared_ptr<FILE> pipe(popen(cmd, "r"), pclose);
	if (!pipe) throw std::runtime_error("popen() failed!");
	while (fgets(buffer.data(), buffer.size(), pipe.get()) != nullptr){
		result += buffer.data();
	}
	return result;
}

void CoviewRobot::getProcessUsage(const std::string& command) {
	std::string cmd = "ps -C " + command + " -o %cpu,%mem --no-headers";
	std::string output = exec(cmd.c_str());

	std::istringstream ss(output);
	float totalCpuUsage = 0.0;
	float totalMemUsage = 0.0;
	float cpuUsage = 0.0;
	float memUsage = 0.0;

	int number = 0;
	while (ss >> cpuUsage >> memUsage) {
		totalCpuUsage += cpuUsage;
		totalMemUsage += memUsage;
		++number;
	}

	if (number > 0) {
		if (command == "test") {
			cpu_total_test = totalCpuUsage;
			mem_total_test = totalMemUsage;
		}
		else if (command == "test_sub") {
			cpu_total_test_sub = totalCpuUsage;
			mem_total_test_sub = totalMemUsage;
		}
	}
}

void CoviewRobot::initPing(){
	try{
		c_ping.clear_access_channels(websocketpp::log::alevel::frame_header);
		c_ping.clear_access_channels(websocketpp::log::alevel::frame_payload);
		c_ping.clear_access_channels(websocketpp::log::alevel::all);

		c_ping.init_asio();

		c_ping.set_open_handler(bind(&CoviewRobot::ping_open, this, ::_1));
		c_ping.set_fail_handler(bind(&CoviewRobot::ping_fail, this, ::_1));
		c_ping.set_message_handler(bind(&CoviewRobot::ping_message, this, ::_1, ::_2));
		c_ping.set_close_handler(bind(&CoviewRobot::ping_close, this, ::_1));

		websocketpp::lib::error_code ec_ping;
		std::ostringstream oss_ping;
		oss_ping << "ws://cobot.center:8286/pang/ws/echo?channel=instant&name=ping2";
		url_ping = oss_ping.str();
		client::connection_ptr con_ping = c_ping.get_connection(url_ping, ec_ping);
		if (!con_ping) {
			std::cerr << "Failed to create connection" << std::endl;
		}
		try{
			hdl_ping = con_ping->get_handle();
			c_ping.connect(con_ping);
		} catch (const websocketpp::exception& e){
			std::cerr << "standart exception: " << e.what() << std::endl;
		}
		c_thread_ping = std::thread(&client::run, &c_ping);
	} catch(...){
		std::cout<<"err initWebSocket()"<<std::endl;
	}
}

void CoviewRobot::initWebSocket(){
	try{
		c.clear_access_channels(websocketpp::log::alevel::frame_header);
		c.clear_access_channels(websocketpp::log::alevel::frame_payload);
		c.clear_access_channels(websocketpp::log::alevel::all);

		c.init_asio();

		c.set_open_handler(bind(&CoviewRobot::on_open, this, ::_1));
		c.set_fail_handler(bind(&CoviewRobot::on_fail, this, ::_1));
		c.set_message_handler(bind(&CoviewRobot::on_message, this, ::_1, ::_2));
		c.set_close_handler(bind(&CoviewRobot::on_close, this, ::_1));

		websocketpp::lib::error_code ec;
		std::ostringstream oss;
		oss << "ws://cobot.center:8286/pang/ws/sub?channel=instant&name=123&track=colink&mode=single&buf_len=3&timeout=4";
		url = oss.str();
		client::connection_ptr con = c.get_connection(url, ec);
		if (!con) {
			std::cerr << "Failed to create connection" << std::endl;
		}
		try{
			hdl = con->get_handle();
			c.connect(con);
		} catch (const websocketpp::exception& e){
			std::cerr << "standart exception: " << e.what() << std::endl;
		}
		c_thread = std::thread(&client::run, &c);
//		c.run();
	} catch(...){
		std::cout<<"err initWebSocket()"<<std::endl;
	}
}

void CoviewRobot::ping_open(websocketpp::connection_hdl hdl){
}

void CoviewRobot::on_open(websocketpp::connection_hdl hdl){
	connection = 1;
}

void CoviewRobot::ping_fail(websocketpp::connection_hdl hdl){
}

void CoviewRobot::on_fail(websocketpp::connection_hdl hdl){
	c.get_alog().write(websocketpp::log::alevel::app, "Connection Failed");
	connection = 0;
	restart_flag = 1;
}

void CoviewRobot::ping_message(websocketpp::connection_hdl hdl, message_ptr msg) {
	try{
		std::vector<uint8_t> payload_ping(msg->get_payload().begin(), msg->get_payload().end());
		uint32_t sent_millis_ping = (payload_ping[0] << 24) | (payload_ping[1] << 16) | (payload_ping[2] << 8) | payload_ping[3];

		auto now_ping = std::chrono::steady_clock::now();
		uint32_t current_millis_ping = std::chrono::duration_cast<std::chrono::milliseconds>(now_ping.time_since_epoch()).count();

		time_ = static_cast<int32_t>(current_millis_ping) - static_cast<int32_t>(sent_millis_ping);
	} catch(...){
	}
}

void CoviewRobot::on_message(websocketpp::connection_hdl hdl, message_ptr msg) {
	try{
		std::vector<uint8_t> ping_time = pingTime();
		std::string payload_ping(ping_time.begin(), ping_time.end());
		c_ping.send(hdl_ping, payload_ping, websocketpp::frame::opcode::binary);

		std::vector<uint8_t> payload(msg->get_payload().begin(), msg->get_payload().end());

		uint32_t sent_millis = (payload[0] << 24) | (payload[1] << 16) | (payload[2] << 8) | payload[3];
		uint8_t now_count = payload[4];

		uint32_t length = 0;
		std::memcpy(&length, payload.data() + 5, sizeof(length));
		size_t dataSize = (payload.size() - 8) * 8;

		bps += dataSize;

		if (count != now_count){
			std::cout << "----------------------------------different start-------------------" << std::endl;
			std::cout << "----------------------------------" << count << ", " << now_count << "-------------------" << std::endl;
			std::cout << "----------------------------------count, now_count-------------------" << std::endl;
			std::cout << "----------------------------------different count-------------------" << std::endl;
			std::cout << "----------------------------------different count-------------------" << std::endl;
			std::cout << "----------------------------------different count-------------------" << std::endl;
			std::cout << "----------------------------------different count-------------------" << std::endl;
			std::cout << "----------------------------------different count-------------------" << std::endl;
			std::cout << "----------------------------------different count-------------------" << std::endl;
			std::cout << "----------------------------------different count-------------------" << std::endl;
			std::cout << "----------------------------------different count-------------------" << std::endl;
			std::cout << "----------------------------------different count-------------------" << std::endl;
			std::cout << "----------------------------------different count-------------------" << std::endl;
			std::cout << "----------------------------------different count-------------------" << std::endl;
			std::cout << "----------------------------------different count-------------------" << std::endl;
			std::cout << "----------------------------------different end-------------------" << std::endl;
			count = now_count;
			if (err_count_flag){
				err_count += 1;
			}
		}
		count++;
		if (count >9){
			count = 0;
		}

		auto now = std::chrono::steady_clock::now();
//		uint32_t current_millis = std::chrono::duration_cast<std::chrono::milliseconds>(now.time_since_epoch()).count();
		auto current_millis = std::chrono::duration_cast<std::chrono::milliseconds>(now.time_since_epoch()).count();

		int32_t time_diff = static_cast<int32_t>(current_millis) - static_cast<int32_t>(sent_millis);

		std::cout << "ERROR Count: " << err_count << ", Ping Delay: " << time_ << " ms, Sent: " << sent_millis << ", Current: " << current_millis << ", Latency(ms): " << time_diff << " ms, Length: " << length << " Byte, Data Size: " << dataSize << " bit, BandWidth: " << std::fixed << std::setprecision(2) << bps_ << "Mbps, CPU_sub: " << std::fixed << std::setprecision(1) << cpu_total_test_sub << " %, Memory_sub: " << mem_total_test_sub << " %, CPU_pub: " << cpu_total_test << " %, Memory_pub: " << mem_total_test << " %."  << std::endl;

		if (file.is_open()){
			writeRow({std::to_string(err_count), std::to_string(time_), std::to_string(sent_millis), std::to_string(current_millis), std::to_string(time_diff), std::to_string(length), std::to_string(dataSize), std::to_string(bps_), std::to_string(cpu_total_test_sub), std::to_string(mem_total_test_sub), std::to_string(cpu_total_test), std::to_string(mem_total_test)});
		}

	} catch (...){
		std::cerr << "Not JSON" << std::endl;
	}
}

void CoviewRobot::ping_close(websocketpp::connection_hdl hdl){
}

void CoviewRobot::on_close(websocketpp::connection_hdl hdl){
	c.get_alog().write(websocketpp::log::alevel::app, "Connection Closed");
	connection = 0;
	restart_flag = 1;
}

int main(int argc, char *argv[]) {
	if (argc < 5) {
		std::cerr << "Usage: " << argv[0] << " <width> <height> <camnum> <bitrate>" << std::endl;
		return 1;
	}

	int width = std::atoi(argv[1]);
	int height = std::atoi(argv[2]);
	int num = std::atoi(argv[3]);
	int bitrate = std::atoi(argv[4]);

	CoviewRobot coview_robot(width, height, num, bitrate);
	std::cout << "_______________________________________________________________________________" << std::endl;
	sleep(3);
	execvp(argv[0], argv);
	exit(EXIT_FAILURE);
	return 0;
}
