#include "CoviewRobot.hpp"


std::vector<uint8_t> pingTime(){
	auto ping_now = std::chrono::steady_clock::now();
	auto millis_ping = std::chrono::duration_cast<std::chrono::milliseconds>(ping_now.time_since_epoch()).count();

	std::vector<uint8_t> result_ping;
	result_ping.resize(3);
	result_ping[0] = (millis_ping >> 16) & 0xFF;
	result_ping[1] = (millis_ping >> 8) & 0xFF;
	result_ping[2] = millis_ping & 0xFF;

	return result_ping;
}

CoviewRobot::CoviewRobot(int w, int h, int n, int b) : width(w), height(h), num(n), bitrate(b)
{
	initVideo();
	initPing();
	initWebSocket();

	while (true){
		if (connection){
			std::this_thread::sleep_for(std::chrono::seconds(1));
			std::vector<uint8_t> ping_time = pingTime();

			std::string payload(ping_time.begin(), ping_time.end());
			c_ping.send(hdl_ping, payload, websocketpp::frame::opcode::binary);
		}
	}
};

void CoviewRobot::initPing(){
	try{
		c_ping.clear_access_channels(websocketpp::log::alevel::frame_header);
		c_ping.clear_access_channels(websocketpp::log::alevel::frame_payload);
		c_ping.clear_access_channels(websocketpp::log::alevel::all);

		c_ping.init_asio();

		c_ping.set_open_handler(bind(&CoviewRobot::ping_open, this, ::_1));
		c_ping.set_fail_handler(bind(&CoviewRobot::ping_fail, this, ::_1));
		c_ping.set_message_handler(bind(&CoviewRobot::ping_message, this, ::_1, ::_2));
		c_ping.set_close_handler(bind(&CoviewRobot::ping_close, this, ::_1));

		websocketpp::lib::error_code ec_ping;
		std::ostringstream oss_ping;
//		oss_ping << "ws://cobiz.cobot.center:8276/pang/ws/pub?channel=cqcs5sj1180c72upie9g&track=video&mode=single&buf_len=3";
//		oss_ping << "ws://192.168.0.11:8276/pang/ws/pub?channel=instant&name=123&track=colink&mode=single&buf_len=3&timeout=4";
		oss_ping << "ws://cobot.center:8286/pang/ws/echo?channel=instant&name=ping";
		url_ping = oss_ping.str();
		client::connection_ptr con_ping = c_ping.get_connection(url_ping, ec_ping);
		if (!con_ping) {
			std::cerr << "Failed to create connection" << std::endl;
		}
		try{
			hdl_ping = con_ping->get_handle();
			c_ping.connect(con_ping);
		} catch (const websocketpp::exception& e){
			std::cerr << "standart exception: " << e.what() << std::endl;
		}
		c_thread_ping = std::thread(&client::run, &c_ping);
//		c.run();
	} catch(...){
		std::cout<<"err initWebSocket()"<<std::endl;
	}
}

void CoviewRobot::initWebSocket(){
	try{
		c.clear_access_channels(websocketpp::log::alevel::frame_header);
		c.clear_access_channels(websocketpp::log::alevel::frame_payload);
		c.clear_access_channels(websocketpp::log::alevel::all);

		c.init_asio();

		c.set_open_handler(bind(&CoviewRobot::on_open, this, ::_1));
		c.set_fail_handler(bind(&CoviewRobot::on_fail, this, ::_1));
		c.set_message_handler(bind(&CoviewRobot::on_message, this, ::_1, ::_2));
		c.set_close_handler(bind(&CoviewRobot::on_close, this, ::_1));

		websocketpp::lib::error_code ec;
		std::ostringstream oss;
//		oss << "ws://cobiz.cobot.center:8276/pang/ws/pub?channel=cqcs5sj1180c72upie9g&track=video&mode=single&buf_len=3";
//		oss << "ws://192.168.0.11:8276/pang/ws/pub?channel=instant&name=123&track=colink&mode=single&buf_len=3&timeout=4";
		oss << "ws://cobot.center:8286/pang/ws/pub?channel=instant&name=123&track=colink&mode=single&buf_len=3&timeout=4";
		url = oss.str();
		client::connection_ptr con = c.get_connection(url, ec);
		if (!con) {
			std::cerr << "Failed to create connection" << std::endl;
		}
		try{
			hdl = con->get_handle();
			c.connect(con);
		} catch (const websocketpp::exception& e){
			std::cerr << "standart exception: " << e.what() << std::endl;
		}
		c_thread = std::thread(&client::run, &c);
//		c.run();
//		c.close(hdl, websocketpp::close::status::normal, "Normal Closure");
	} catch(...){
		std::cout<<"err initWebSocket()"<<std::endl;
	}
//	std::cout<<"init websocket"<<std::endl;
}

void CoviewRobot::ping_open(websocketpp::connection_hdl hdl){
}

void CoviewRobot::on_open(websocketpp::connection_hdl hdl){
	connection = 1;
	std::string msg = "";
	std::ostringstream mime;
//	mime << "image/jpeg;width=" << width << ";height=" << height << ";framerate=30;codecs=jpeg";
	mime << "video/h264;width=" << width << ";height=" << height << ";framerate=30;codecs=avc1.42002A";
	msg = mime.str();
//	msg = "video/h264;width=1920;height=1080;framerate=30;codecs=avc1.42002A";
	c.send(hdl, msg, websocketpp::frame::opcode::text);
	c.get_alog().write(websocketpp::log::alevel::app, "Sent Message: " + msg);
}

void CoviewRobot::ping_fail(websocketpp::connection_hdl hdl){
}

void CoviewRobot::on_fail(websocketpp::connection_hdl hdl){
	c.get_alog().write(websocketpp::log::alevel::app, "Connection Failed");
	connection = 0;
	restart_flag = 1;
}

void CoviewRobot::ping_message(websocketpp::connection_hdl hdl, message_ptr msg) {
	try{
		std::vector<uint8_t> payload(msg->get_payload().begin(), msg->get_payload().end());
		uint8_t sent_millis = (payload[0] << 16) | (payload[1] << 8) | payload[2];

		auto now = std::chrono::steady_clock::now();
		uint8_t current_millis = std::chrono::duration_cast<std::chrono::milliseconds>(now.time_since_epoch()).count();
		time_ = static_cast<int32_t>(current_millis) - static_cast<int32_t>(sent_millis);

		std::cout << "Ping Delay: " << time_ << " ms" << std::endl;
	} catch(...){

	}
}

void CoviewRobot::on_message(websocketpp::connection_hdl hdl, message_ptr msg) {
	try{
		json jsonData = json::parse(msg->get_payload());
		c.get_alog().write(websocketpp::log::alevel::app, "Received Reply: " + msg->get_payload());
	} catch (...){
		std::cerr << "Not JSON" << std::endl;
	}
}

void CoviewRobot::ping_close(websocketpp::connection_hdl hdl){
}

void CoviewRobot::on_close(websocketpp::connection_hdl hdl){
	c.get_alog().write(websocketpp::log::alevel::app, "Connection Closed");
	connection = 0;
	restart_flag = 1;
}

std::vector<uint8_t> addHeaderToGstMapData(uint8_t count, const GstMapInfo& map) {
	auto now = std::chrono::steady_clock::now();
	auto millis = std::chrono::duration_cast<std::chrono::milliseconds>(now.time_since_epoch()).count();

	uint32_t length = (uint32_t)map.size;

	std::vector<uint8_t> result;
	result.resize(9);
	result[0] = (millis >> 24) & 0xFF;
	result[1] = (millis >> 16) & 0xFF;
	result[2] = (millis >> 8) & 0xFF;
	result[3] = millis & 0xFF;
	result[4] = count;

	result.resize(8);
	std::memcpy(result.data() + 5, &length, sizeof(length));

	uint32_t sent_millis = (result[0] << 24) | (result[1] << 16) | (result[2] << 8) | result[3];
	std::cout << "Sent milliseconds: " << sent_millis << std::endl;

	result.insert(result.end(), map.data, map.data + map.size);

	return result;
}

GstFlowReturn CoviewRobot::on_new_sample(GstAppSink *appsink, gpointer user_data){
	CoviewRobot* instance = static_cast<CoviewRobot*>(user_data);
	GstSample* sample = gst_app_sink_pull_sample(GST_APP_SINK(appsink));
	if (sample != NULL){
		try {
			GstBuffer *buffer = gst_sample_get_buffer(sample);
			GstMapInfo map;
			if (gst_buffer_map(buffer, &map, GST_MAP_READ)) {
				std::vector<uint8_t> result = addHeaderToGstMapData(instance->count, map);
				instance->count += 1;
				if (instance->count > 9){
					instance->count = 0;
				}
//				std::vector<unsigned char> frameData(map.data, map.data + map.size);

//				frameData.insert(frameData.begin(), instance->fourcc_high, instance->fourcc_high + 4);
//				std::cout << "Data with 4CC added. Size: " << frameData.size() << ", Normal Size: " << map.size << std::endl;
				std::string payload(result.begin(), result.end());
				if (instance->connection) {
					instance->c.send(instance->hdl, payload, websocketpp::frame::opcode::binary);
//					instance->c.send(instance->hdl, frameData.data(), frameData.size(), websocketpp::frame::opcode::binary);
//					instance->c.send(instance->hdl, map.data, map.size, websocketpp::frame::opcode::binary);
				}
				gst_buffer_unmap(buffer, &map);
			}
			gst_sample_unref(sample);
		} catch (const websocketpp::exception& e) {
			std::cerr << "Err: " << e.what() << std::endl;
			gst_sample_unref(sample);
		}
	}
	return GST_FLOW_OK;
}

void CoviewRobot::initVideo() {
	const char *pipeline_cmd = "";
	std::ostringstream text;
	text << "v4l2src device=/dev/video" << num << " ! video/x-raw, width=" << width << ", height=" << height << ", framerate=30/1 ! videoconvert ! x264enc bitrate="<< bitrate << " key-int-max=30 speed-preset=ultrafast tune=zerolatency ! video/x-h264, alignment=au, stream-format=byte-stream ! h264parse config-interval=1 ! appsink name=sink drop=true emit-signals=true max-buffers=3 sync=false";
//	text << "v4l2src device=/dev/video" << num << " ! video/x-raw, width=" << width << ", height=" << height << ", framerate=30/1 ! tee name=t t. ! queue ! videoconvert ! x264enc bitrate="<< bitrate << " key-int-max=30 speed-preset=ultrafast tune=zerolatency ! video/x-h264, alignment=au, stream-format=byte-stream ! h264parse config-interval=1 ! appsink name=sink drop=true emit-signals=true max-buffers=3 sync=false t. ! queue ! videoscale ! video/x-raw, width=1280, height=720 ! videoconvert ! x264enc bitrate=4096 key-int-max=30 speed-preset=ultrafast tune=zerolatency ! video/x-h264, alignment=au, stream-format=byte-stream ! h264parse config-interval=1 ! appsink name=sink_mid drop=true emit-signals=true max-buffers=3 sync=false t. ! queue ! videoscale ! video/x-raw, width=640, height=480 ! videoconvert ! x264enc bitrate="<< bitrate << " key-int-max=30 speed-preset=ultrafast tune=zerolatency ! video/x-h264, alignment=au, stream-format=byte-stream ! h264parse config-interval=1 ! appsink name=sink_low drop=true emit-signals=true max-buffers=3 sync=false";
	std::string text_str = text.str();
	pipeline_cmd = text_str.c_str();
	pipeline = gst_parse_launch(pipeline_cmd, NULL);
	appsink = gst_bin_get_by_name(GST_BIN(pipeline), "sink");
//	appsink_mid = gst_bin_get_by_name(GST_BIN(pipeline), "sink_mid");
//	appsink_low = gst_bin_get_by_name(GST_BIN(pipeline), "sink_low");
	g_signal_connect(appsink, "new-sample", G_CALLBACK(CoviewRobot::on_new_sample), this);
//	g_signal_connect(appsink_mid, "new-sample", G_CALLBACK(CoviewRobot::on_new_sample_mid), this);
//	g_signal_connect(appsink_low, "new-sample", G_CALLBACK(CoviewRobot::on_new_sample_low), this);
	gst_element_set_state(pipeline, GST_STATE_PLAYING);
	std::cout<<"Video State PLAYING"<<std::endl;
}

void CoviewRobot::closeVideo() {
	gst_element_set_state(pipeline, GST_STATE_NULL);
}

int main(int argc, char *argv[]) {
	gst_init(&argc, &argv);
	if (argc < 5) {
		std::cerr << "Usage: " << argv[0] << " <width> <height> <camnum> <bitrate>" << std::endl;
		return 1;
	}
	int width = std::atoi(argv[1]);
	int height = std::atoi(argv[2]);
	int num = std::atoi(argv[3]);
	int bitrate = std::atoi(argv[4]);

	CoviewRobot coview_robot(width, height, num, bitrate);
	std::cout << "_______________________________________________________________________________" << std::endl;
	coview_robot.closeVideo();
	sleep(3);
	execvp(argv[0], argv);
	exit(EXIT_FAILURE);
	return 0;
}
